
/*
 * This contains roughly the same structure as the main decoder,
 * except that instead of executing the instruction, we write 
 * an assembler-like representation of it to a passed buffer.
 *
 */

#include "sim.h"


/* [row][col][stringlen] */
static char regs[16][4][15] = {
  {"ax0", "i0", "i4", "astat"},
  {"ax1", "i1", "i5", "mstat"},
  {"mx0", "i2", "i6", "sstat"},
  {"mx1", "i3", "i7", "imask"},
  {"ay0", "m0", "m4", "icntl"},
  {"ay1", "m1", "m5", "cntr"},
  {"my0", "m2", "m6", "sb"},
  {"my1", "m3", "m7", "px"},
  {"si", "l0", "l4", "rx0"},
  {"se", "l1", "l5", "tx0"},
  {"ar", "l2", "l6", "rx1"},
  {"mr0", "l3", "l7", "rx1"},
  {"mr1", "", "", "ifc"},
  {"mr2", "", "", "owrcntr"},
  {"sr0", "pmovlay", "", ""},
  {"sr1", "", "", ""}
};

static char xops[8][15] = {
  "x0",
  "x1",
  "ar",
  "mr0",
  "mr1",
  "mr2",
  "sr0",
  "sr1"
};

static char yops[4][15] = {
  "y0",
  "y1",
  "f",
  "zero"
};

static char sfs[16][3][15] = {
  {"lshift", "(hi)", {0}},
  {"lshift", "(hi)", {1}},
  {"lshift", "(lo)", {0}},
  {"lshift", "(lo)", {1}},
  {"ashift", "(hi)", {0}},
  {"ashift", "(hi)", {1}},
  {"ashift", "(lo)", {0}},
  {"ashift", "(lo)", {1}},
  {"norm", "(hi)", {0}},
  {"norm", "(hi)", {1}},
  {"norm", "(lo)", {0}},
  {"norm", "(lo)", {1}},
  {"exp", "(hi)", {0}},
  {"exp", "(hix)", {0}},
  {"exp", "(lo)", {0}},
  {"dbe", "", {0}}
};

static char conds[16][15] = {
  "eq",
  "ne", 
  "gt", 
  "le",
  "lt",
  "ge",
  "av",
  "not av",
  "ac", 
  "not ac",
  "neg",
  "pos",
  "mv",
  "not mv",
  "not ce",
  "" /* always true */
};

static char flagops[4][15] = {
  "nochange",
  "toggle",
  "reset",
  "set"
};

/*
 * Returns the reg/rgp for the register name (of len namelen).
 *
 */
int findreg(char *name, int namelen, int *reg, int *rgp)
{
  int i,j;
  if (!name || (!reg || !rgp))
    return 0;

  for (i=0;i<16;i++)
    {
      for (j=0;j<4;j++)
	{
	  if (!strncmp(name, regs[i][j], namelen))
	    {
	      *reg = i;
	      *rgp = j;
	      return 1;
	    }
	}
    }
  
  return 0;
}

unsigned short getconstant(unsigned int yy, unsigned int cc, unsigned int bo)
{
  if(bo==1)
    {
      switch(yy)
	{
	case 0x00:
	  switch(cc)
	    {
	    case 0x00: return 0x0001;
	    case 0x01: return 0x0002;
	    case 0x02: return 0x0004;
	    case 0x03: return 0x0008;
	    }
	case 0x01:
	  switch(cc)
	    {
	    case 0x00: return 0x0010;
	    case 0x01: return 0x0020;
	    case 0x02: return 0x0040;
	    case 0x03: return 0x0080;
	    }
	case 0x02:
	  switch(cc)
	    {
	    case 0x00: return 0x0100;
	    case 0x01: return 0x0200;
	    case 0x02: return 0x0400;
	    case 0x03: return 0x0800;
	    }
	case 0x03:
	  switch(cc)
	    {
	    case 0x00: return 0x1000;
	    case 0x01: return 0x2000;
	    case 0x02: return 0x4000;
	    case 0x03: return 0x8000;
	    }
	}
    }
  else if (bo==3)
    {
      switch(yy)
	{
	case 0x00:
	  switch(cc)
	    {
	    case 0x00: return 0xfffe;
	    case 0x01: return 0xfffd;
	    case 0x02: return 0xfffb;
	    case 0x03: return 0xfff7;
	    }
	case 0x01:
	  switch(cc)
	    {
	    case 0x00: return 0xffef;
	    case 0x01: return 0xffdf;
	    case 0x02: return 0xffbf;
	    case 0x03: return 0xff7f;
	    }
	case 0x02:
	  switch(cc)
	    {
	    case 0x00: return 0xfeff;
	    case 0x01: return 0xfdff;
	    case 0x02: return 0xfbff;
	    case 0x03: return 0xf7ff;
	    }
	case 0x03:
	  switch(cc)
	    {
	    case 0x00: return 0xefff;
	    case 0x01: return 0xdfff;
	    case 0x02: return 0xbfff;
	    case 0x03: return 0x7fff;
	    }
	}
    }
  else
    fprintf(stderr, "pretty: unknown bo\n");
  return 0;
}

int prettydecode(unsigned int opcode, char *buf)
{
  if (!buf)
    return -1;

  memset(buf, 0x00, MAXDECODE+1);

  /* No-oper codes... */
  if ((opcode >> 16) == 0x0000)
    {
      sprintf(buf, "nop");
      return 0;
    }
  if ((opcode>>4) == 0x02800)
    {
      if (opcode & 0xf)
	sprintf(buf, "idle(%d)", opcode & 0xf);
      else
	sprintf(buf, "idle");
      return 0;
    }

  /* full codes... */
  if (opcode & BIT23) /* 1 */
    {
      if (opcode & BIT22) /* 11 */
	{
	  if (opcode & BIT21) /* 111 */
	    {
	      sprintf(buf, "111");
	    }
	  else /* 110 */
	    {
	      sprintf(buf, "110");
	    }
	}
      else /* 10 */
	{
	  if (opcode & BIT21) /* 101 */
	    {
	   
	    }
	  else /* 100 */
	    {

	    }
	}
    }
  else /* BIT23 off */
    {
      if (opcode & BIT22) /* 01 */
	{
	  if (opcode & BIT21) /* 011 */
	    {

	    }
	  else /* 010 */
	    {
	      if (opcode & BIT20) /* 0101 */
		{
		  if (opcode & BIT19) /* 0101 1 */
		    {
		      if (opcode & BIT18) /* 0101 11 */
			{
			  sprintf(buf, "0101 11");
			}
		      else /* 0101 10 */
			{
#if 0
			  unsigned int amf = (opcode >> 13) & 0x1f;
			  unsigned int dreg = (opcode >> 4) & 0xf;
			  unsigned int i = (opcode >> 1) & 0x1;
			  unsigned int m = opcode & 0x1;
#endif
			  sprintf(buf, "0101 10");
 			}
		    }
		  else /* 01010 */
		    {
		      sprintf(buf, "01010");
		    }
		}
	      else /* 0100 */
		{

		}
	    }
	}
      else  /* 00 */
	{
	  if (opcode & BIT21) /* 001 */
	    {
	      if (opcode & BIT20) /* 0011 */
		{
	
		}
	      else /* 0010 */
		{
		  if (opcode & BIT19) /* 0010 1 */
		    {
		
		    }
		  else /* 0010 0 */
		    {
	
		    }
		}
	    }
	  else /* 000 */
	    {
	      if (opcode & BIT20) /* 0001 */
		{
		  if (opcode & BIT19) /* 00011 */
		    {

		    }
		  else /* 00010 */
		    {
		      if (opcode & BIT18)
			{
			}
		      else
			sprintf(buf, "000101");
		    }
		}
	      else /* 0000 */
		{
		  if (opcode & BIT19) /* 0000 1 */
		    {
		      if (opcode & BIT18) /* 0000 11 */
			{
			  if (opcode & BIT17) /* 0000 111 */
			    {
			      if (opcode & BIT16) /* 0000 1111 */
				{
				}
			      else /* 0000 1110 */
				{
				  sprintf(buf, "0000 1110");
				}
			    }
			  else /* 0000 110 */
			    {
			      if (opcode & BIT16)
				{
			
				}
			      else /* 0000 1100 */
				{
	
				}
				
			    }
			}
		      else /* 0000 10 */
			{
		
			  else if ((opcode >> 4) == 0x09080)
			    {
		
			    }
			  else if ((opcode>>8) == 0xb00)
			    {
			      sprintf(buf, "%s%s %s (i%d)",
				      (opcode&0xf)?"if ":"",
				      (opcode&0xf)?conds[(opcode&0xf)]:"",
				      ((opcode>>4)&0x1)?"call":"jump",
				      (opcode>>6)&0x8);
			    }
			}
		    }
		  else /* 0000 0 */
		    {
		      if (opcode & BIT18) /* 0000 01 */
			{
			  if ((opcode >> 5) == 0x2000)
			    {
			      /* stack control */
			      sprintf(buf, "[stack control]");
			    }

			  if (opcode & BIT17) /* 0000 011 */
			    {
			      if (opcode & BIT16) /* 0000 0111 */
				{
				  /* DIVQ */
				}
			      else /* 0000 0110 */
				{
				  /* DIVS */

				}
			    }
			  else /* 0000 010 */
			    {
			      sprintf(buf, "0000 010");
			    }
			}
		      else /* 000000 */
			{
			  if (opcode & BIT17) /* 0000 001 */
			    {
			      if (opcode & BIT16) /* 0000 0011 */
				{
				  /* call/jump on flag in */

				}
			      else /* 0000 0010 */
				{
				  /* modify flag out */

				}
			    }
			  else /* 0000 000 */
			    {
			      if (opcode & BIT16) /* 0000 0001 */
				{
		
				}
			      else /* 0000 0000 */
				{
				  sprintf(buf, "0000 0000");
				}
			    }
			}
		    }
		}
	    }
	}
    }
  return 0;
}
