
#include "sim.h"
#include "pretty_tables.h"

/*
 * Type 1: ALU/MAC with Data and Program Memory Read
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  1 -PD-- -DD-- ----AMF------- -Yop- -Xop---- -PMI- -PMM- -DMI- -DMM-
 *
 */
unsigned int prettyop_pmdmr(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_pmdmr]");
  return 0;
}

/*
 * Type 2: Data Memory Write (Immediate Data)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  0  1 -G ------------------------ADDR------------------- --I-- --M--
 *
 */
unsigned int prettyop_dmwim(unsigned int opcode, char *buf)
{
  unsigned int dag = (opcode >> 20)&0x1;
  unsigned int data = (opcode >> 4)&0x1ffff;
  unsigned int i = (opcode >> 2)&0x3;
  unsigned int m = opcode&0x3;
  sprintf(buf, "dm(i%d,m%d) = 0x%04x", (!dag)?i:i+4, (!dag)?m:m+4, data);
  return 0;
}

/*
 * Type 3: Read/Write Data Memory (Immediate Address)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  1  0  0 -D ---RGP-- ---------------ADDR------------------- ----REG----
 *
 */
unsigned int prettyop_dmrwim(unsigned int opcode, char *buf)
{
  if (opcode & BIT20) /* 1001 */
    {
      unsigned int addr = (opcode >> 4) & 0x3fff;
      unsigned int rgp = (opcode >> 18) & 0x3;
      unsigned int reg = opcode & 0xf;
      sprintf(buf, "dm(0x%04x) = %s", addr, regs[reg][rgp]);
    }
  else /* BIT20 off */
    {
      unsigned int addr = (opcode >> 4) & 0x3fff;
      unsigned int rgp = (opcode >> 18) & 0x3;
      unsigned int reg = opcode & 0xf;
      sprintf(buf, "%s = dm(0x%04x)", regs[reg][rgp], addr);
    }
  return 0;
}

/*
 * Type 4: ALU/MAC with Data Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  1  G  D  Z  -----AMF----  -Yop- --Xop---  ---DREG--- --I-- --M--
 *
 * ALU/MAC Optional.
 *
 */
unsigned int prettyop_dmrw(unsigned int opcode, char *buf)
{
  /* dm(i, m) = dreg */
  unsigned int dag = (opcode & BIT20)?1:0;
  if ((opcode & BIT19) && !(opcode & BIT18))
    {
      unsigned int amf = (opcode >> 13) & 0x1f;
      unsigned int i = (opcode >> 2) & 0x3;
      unsigned int m = opcode & 0x3;
      unsigned int dreg = (opcode >> 4) & 0xf;
      sprintf(buf, "dm(i%d, m%d) = %s %s", dag?i+4:i, dag?m+4:m, regs[dreg][0], amf?"(AMF unhandled)":"");
    }
  else /* no amf */
    {
      unsigned int data = (opcode >> 4) & 0xff;
      unsigned int i = (opcode >> 1) & 0x1;
      unsigned int m = opcode & 0x1;
      sprintf(buf, "dm(i%d, m%d) = 0x%02x", i, m, data); 
    }
  return 0;
}

/*
 * Type 5: ALU/MAC with Program Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  0  1 -D -Z -----AMF------ -Yop- --Xop--- ----DREG--- --I-- --M--
 *
 * ALU/MAC optional.
 *
 */
unsigned int prettyop_pmrw(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_pmrw]");
  return 0;
}

/*
 * Type 6: Load Data Register Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  1  0  0 -----------------------DATA-------------------- ----DREG---
 *
 */
unsigned int prettyop_loadregim(unsigned int opcode, char *buf)
{
  unsigned int addr = (opcode >> 4) & 0x3fff;
  unsigned int dreg = opcode & 0xf;
  sprintf(buf, "%s = 0x%04x", regs[dreg][0], addr);
  return 0;
}

/*
 * Type 7: Load Non-Data Register Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  1 -RGP- -----------------DATA-------------------- ----REG----
 *
 */
unsigned int prettyop_loadnondataregim(unsigned int opcode, char *buf)
{
  unsigned int addr = (opcode >> 4) & 0x3fff;
  unsigned int rgp = (opcode >> 18) & 0x3;
  unsigned int reg = opcode & 0xf;
  sprintf(buf, "%s = 0x%04x", regs[reg][rgp], addr);
  return 0;
}

/*
 * Type 8a: ALU/MAC with Internal Data Register Move 
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  1 -Z -----AMF------ -Yop- ---Xop--  -DestDREG-  --SrcDREG-
 *
 *
 * Type 8b: Generate ALU Status (none = <ALU>)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  1  0 -----AMF------ -Yop- ---Xop--  1  0  1  0  1  0  1  0
 *                      |
 *                       -- ALU codes only (no MAC) 
 *
 */
unsigned int prettyop_alumove(unsigned int opcode, char *buf)
{
  unsigned int z = (opcode >> 18) & 0x1;
  unsigned int amf = (opcode >> 13) & 0x1f;
  unsigned int yop = (opcode >> 11) & 0x3;
  unsigned int xop = (opcode >> 8) & 0x3;
#if 0
  unsigned int destdreg = (opcode >> 4) & 0xf;
  unsigned int srcdreg = opcode & 0xf;
#endif
  sprintf(buf, "%s = ", z?"af":"ar");
  switch(amf)
    {
    case 0x10:
      sprintf(buf+strlen(buf), "%s", yop?ayops[yop]:""); break;
    case 0x11:
      if (yop)
	sprintf(buf+strlen(buf), "%s + 1", ayops[yop]);
      else
	sprintf(buf+strlen(buf), "pass 1");
      break;
    case 0x12:
      sprintf(buf+strlen(buf), "%s + %s + c", axops[xop], ayops[yop]);
      break;
    case 0x13:
      if (yop)
	sprintf(buf+strlen(buf), "%s + %s", axops[xop], axops[yop]);
      else
	sprintf(buf+strlen(buf), "%s", axops[xop]);
      break;
    default:
      fprintf(stderr, "decode: unhandled amf %02x\n", amf);
    }
  return 0;
}

int prettydecode_amf(char *buf, unsigned int z, unsigned int amf, unsigned int xop, unsigned int yop, unsigned int cc, unsigned int bo)
{
  if (amf>0xf)
    sprintf(buf+strlen(buf), "%s = ", z?"af":"ar");
  else
    sprintf(buf+strlen(buf), "%s = ", z?"mf":"mr");
  
  switch(amf)
    {
      /*
       * MAC
       */
    case 0x01:
      sprintf(buf+strlen(buf), "%s * %s (rnd)", mxops[xop], myops[yop]);
      break;
    case 0x02:
      sprintf(buf+strlen(buf), "mr + %s * %s (rnd)", mxops[xop], myops[yop]);
      break;
    case 0x03:
      sprintf(buf+strlen(buf), "mr - %s * %s (rnd)", mxops[xop], myops[yop]);
      break;  
    case 0x04:
      if (yop == 0x3)
	sprintf(buf+strlen(buf), "0"); /* CLEAR */
      else
	sprintf(buf+strlen(buf), "%s * %s (ss)", mxops[xop], myops[yop]);
      break;
    case 0x05:
      sprintf(buf+strlen(buf), "%s * %s (su)", mxops[xop], myops[yop]);
      break;
    case 0x06:
      sprintf(buf+strlen(buf), "%s * %s (us)", mxops[xop], myops[yop]);
      break;
    case 0x07:
      sprintf(buf+strlen(buf), "%s * %s (uu)", mxops[xop], myops[yop]);
      break;
    case 0x08:
      sprintf(buf+strlen(buf), "mr + %s * %s (ss)", mxops[xop], myops[yop]);
      break;
    case 0x09:
      sprintf(buf+strlen(buf), "mr + %s * %s (su)", mxops[xop], myops[yop]);
      break;
    case 0x0a:
      sprintf(buf+strlen(buf), "mr + %s * %s (us)", mxops[xop], myops[yop]);
      break;
    case 0x0b:
      sprintf(buf+strlen(buf), "mr + %s * %s (uu)", mxops[xop], myops[yop]);
      break;
    case 0x0c:
      sprintf(buf+strlen(buf), "mr - %s * %s (ss)", mxops[xop], myops[yop]);
      break;
    case 0x0d:
      sprintf(buf+strlen(buf), "mr - %s * %s (su)", mxops[xop], myops[yop]);
      break;
    case 0x0e:
      sprintf(buf+strlen(buf), "mr - %s * %s (us)", mxops[xop], myops[yop]);
      break;
    case 0x0f:
      sprintf(buf+strlen(buf), "mr - %s * %s (uu)", mxops[xop], myops[yop]);
      break;

      /*
       * ALU
       */
    case 0x10:
      if (yop==0x3)
	sprintf(buf+strlen(buf), "%s", ayops[yop]);
      else
	sprintf(buf+strlen(buf), "pass %s", ayops[yop]);
      break;
    case 0x11:
      if (yop==0x3)
	sprintf(buf+strlen(buf), "pass 1");
      else
	sprintf(buf+strlen(buf), "%s + 1", ayops[yop]);
      break;
    case 0x12:
      sprintf(buf+strlen(buf), "%s + %s + c", axops[xop], ayops[yop]);
      break;
    case 0x13:
      if (yop == 0x3)
	sprintf(buf+strlen(buf), "pass %s", axops[xop]);
      else
	sprintf(buf+strlen(buf), "%s + %s", axops[xop], ayops[yop]);
      break;
    case 0x14:
      sprintf(buf+strlen(buf), "not %s", ayops[yop]);
      break;
    case 0x15:
      sprintf(buf+strlen(buf), "-%s", ayops[yop]);
      break;
    case 0x16:
      if(yop==0x3)
	sprintf(buf+strlen(buf), "%s + c - 1", axops[xop]);
      else
	sprintf(buf+strlen(buf), "%s - %s + c - 1", axops[xop], ayops[yop]);
      break;
    case 0x17:
      if (!cc && !bo)
	sprintf(buf+strlen(buf), "%s - %s", axops[xop], ayops[yop]);
      else
	sprintf(buf+strlen(buf), "%s - %d", axops[xop], getconstant(yop, cc, bo));
      break;
    case 0x18:
      if (yop == 0x3)
	sprintf(buf+strlen(buf), "pass - 1");
      else
	sprintf(buf+strlen(buf), "%s - 1", ayops[yop]);
      break;
    case 0x19:
      if (yop == 0x3)
	sprintf(buf+strlen(buf), "-%s", axops[xop]);
      else
	sprintf(buf+strlen(buf), "%s - %s", ayops[yop], axops[xop]);
      break;
    case 0x1a:
      if (yop == 0x3)
	sprintf(buf+strlen(buf), "-%s + c - 1", axops[xop]);
      else
	sprintf(buf+strlen(buf), "%s - %s + c - 1", ayops[yop], axops[xop]);
      break;
    case 0x1b:
      sprintf(buf+strlen(buf), "not %s", axops[xop]);
      break;
    case 0x1c:
      sprintf(buf+strlen(buf), "%s and %s", axops[xop], ayops[yop]);
      break;
    case 0x1d:
      sprintf(buf+strlen(buf), "%s or %s", axops[xop], ayops[yop]);
      break;
    case 0x1e:
      sprintf(buf+strlen(buf), "%s xor %s", axops[xop], ayops[yop]);
      break;
    case 0x1f:
      sprintf(buf+strlen(buf), "abs %s", axops[xop]);
      break;
    default:
      sprintf(buf+strlen(buf), "[amf=0x%04x]", amf);
      break;
    }
  return 0;
}

/*
 * Type 9's... Conditional ALU/MAC
 *
 * Type 9a: xop * yop
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -Yop- ---Xop--  0  0  0  0 ---COND----
 *
 * Type 9b: xop * xop
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------  0  0 ---Xop--  0  0  0  1 ---COND----
 *
 * Type 9c: xop {AND,OR,XOR} constant
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -YY-- ---Xop-- -CC-- -BO-- ---COND----
 *
 * Type 9d: PASS constant (constant != 0,1,-1)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  1  0  0 -Z -----AMF------ -YY-- ---Xop-- -CC-- -BO-- ---COND----
 *
 */
unsigned int prettyop_alu(unsigned int opcode, char *buf)
{
  unsigned int z = (opcode >> 18) & 0x1;
  unsigned int amf = (opcode >> 13) & 0x1f;
  unsigned int yop = (opcode >> 11) & 0x3;
  unsigned int xop = (opcode >> 8) & 0x7;
  unsigned int cc = (opcode >> 6) & 0x3;
  unsigned int bo = (opcode >> 4) & 0x3;
  unsigned int cond = opcode & 0xf;

  buf[0] = '\0';
  
  if (cond<0xf)
    sprintf(buf+strlen(buf), "if %s ", conds[cond]);

  prettydecode_amf(buf, z, amf, xop, yop, cc, bo);
  return 0;
}

/*
 * Type 10: Conditional Jump (Immediate Addres)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  1 -S ------------------ADDR------------------- ----COND---
 *
 */
unsigned int prettyop_jump(unsigned int opcode, char *buf)
{	      
  if (opcode & BIT18) /* 000111 */
    {			  
      unsigned int cond = opcode & 0xf;
      unsigned int addr = (opcode >> 4) & 0x3fff;
      if (cond < 0xf)
	sprintf(buf, "if %s call 0x%04x", conds[cond], addr);
      else
	sprintf(buf, "call 0x%04x", addr);
    }
  else /* 000110 */
    {
      unsigned int cond = opcode & 0xf;
      unsigned int addr = (opcode >> 4) & 0x3fff;
      if (cond < 0xf)
	sprintf(buf, "if %s jump 0x%04x", conds[cond], addr);
      else
	sprintf(buf, "jump 0x%04x", addr);
    }
  return 0;
}

/*
 * Type 11: Do Until
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  1 --------------------ADDR----------------- ----COND---
 *
 */
unsigned int prettyop_dountil(unsigned int opcode, char *buf)
{
  unsigned short addr,cond;
  addr = (opcode>>4) & 0x3fff;
  cond = opcode & 0xf;
  sprintf(buf, "do %04x until %s", addr, loopconds[cond]);
  return 0;
}

/*
 * Type 12: Shift with Data Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  1 -G -D ----SF----- --Xop--- ---Dreg---- --I-- --M--
 *
 */
unsigned int prettyop_shiftdmrw(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_shiftdmrw]");
  return 0;
}

/*
 * Type 13: Shift with Program Memory Read/Write
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  0  1 -D ----SF----- --Xop--- ---Dreg---- --I-- --M--
 *
 */
unsigned int prettyop_shiftpmrw(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_shiftpmrw]");
  return 0;
}

/*
 * Type 14: Shift with Internal Data Register Move
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  1  0  0  0  0  0 ----SF----- --Xop--- --DestDREG- --SrcDREG--
 *
 */
unsigned int prettyop_shiftmove(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_shiftmove]");
  return 0;
}

/*
 * Type 15: Shift Immediate
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  1  1  0 ----SF----- --XOP--- --------exponent-------
 *
 */
unsigned int prettyop_shiftim(unsigned int opcode, char *buf)
{
  unsigned int sf = (opcode >> 11) & 0xf;
  unsigned int xop = (opcode >> 8) & 0x7;
  unsigned int exp = opcode & 0xff;
  sprintf(buf, "sr = %s %s %s by 0x%04x %s", 
	  sfs[sf][2][0]==1?"sr or":"",
	  sfs[sf][0],
	  axops[xop],
	  exp,
	  sfs[sf][1]);
  return 0;
}

/*
 * Type 16: Conditional Shift
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  1  0  0 ----SF----- ---XOP--  0  0  0  0 ----COND---
 *
 */
unsigned int prettyop_shift(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_shift]");
  return 0;
}

/*
 * Type 17: Internal Data Move
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  0  1  0  0  0  0 -DRGP -SRGP --DestReg-- ---SrcReg--
 *
 */
unsigned int prettyop_datamove(unsigned int opcode, char *buf)
{	
  sprintf(buf, "%s = %s", 
	  regs[(opcode>>4)&0xf][(opcode>>10)&0x3], 
	  regs[opcode&0xf][(opcode>>8)&0x3]);
  return 0;
}

/*
 * Type 18: Mode Control
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  1  0  0 -TI-- -MM-- -AS-- -OL-- -BR-- -SR-- -GM--  0  0
 *
 */
unsigned int prettyop_modectl(unsigned int opcode, char *buf)
{	
  buf[0] = '\0';
  
  if (((opcode>>2)&03)>=2)
    {
      if (((opcode>>2)&0x3)==2)
	strcat(buf, "dis g_mode ");
      else
	strcat(buf, "ena g_mode ");
    }
  
  if (((opcode>>4)&03)>=2)
    {
      if (((opcode>>4)&0x3)==2)
	strcat(buf, "dis sec_reg ");
      else
	strcat(buf, "ena sec_reg ");
    }
  
  if (((opcode>>6)&03)>=2)
    {
      if (((opcode>>6)&0x3)==2)
	strcat(buf, "dis bit_rev ");
      else
	strcat(buf, "ena bit_rev ");
    }
  
  if (((opcode>>8)&03)>=2)
    {
      if (((opcode>>8)&0x3)==2)
	strcat(buf, "dis av_latch ");
      else
	strcat(buf, "ena av_latch ");
    }
  
  if (((opcode>>0x0a)&03)>=2)
    {
      if (((opcode>>0x0a)&0x3)==2)
	strcat(buf, "dis ar_sat ");
      else
	strcat(buf, "ena ar_sat ");
    }
  
  if (((opcode>>0x0c)&03)>=2)
    {
      if (((opcode>>0x0c)&0x3)==2)
	strcat(buf, "dis m_mode ");
      else
	strcat(buf, "ena m_mode ");
    }
  
  if (((opcode>>0x0e)&03)>=2)
    {
      if (((opcode>>0x0e)&0x3)==2)
	strcat(buf, "dis timer ");
      else
	strcat(buf, "ena timer ");
    }
  return 0;
}

/*
 * Type 19: Conditional Jump (Indirect Address)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  1  1  0  0  0  0  0  0  0  0 --I--  0 -S ----COND---
 *
 */
unsigned int prettyop_jumpindirect(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_jumpindirect]");
  return 0;
}

/*
 * Type 20: Conditional Return 
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0 -T ----COND---
 *
 */
unsigned int prettyop_return(unsigned int opcode, char *buf)
{
  if ((opcode >> 4) == 0xa001) /* rti */
    {
      unsigned int cond = opcode & 0xf;
      if (cond < 0xf)
	sprintf(buf, "if %s rti", conds[cond]);
      else
	sprintf(buf, "rti");
    }
  else if ((opcode >> 4) == 0xa000) /* rts */
    {
      unsigned int cond = opcode & 0xf;
      if (cond < 0xf)
	sprintf(buf, "if %s rts", conds[cond]);
      else
	sprintf(buf, "rts");
    }
  return 0;
}

/*
 * Type 21: Modify Address Register
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  0  1  0  0  0  0  0  0  0  0  0  0  0 -G --I-- --M--
 *
 */
unsigned int prettyop_modify(unsigned int opcode, char *buf)
{
  unsigned int dag = (opcode >> 4) & 0x1;
  unsigned int i = (opcode >> 2) & 0x3;
  unsigned int m = opcode & 0x3;
  sprintf(buf, "modify (i%d, m%d)", dag?i+4:i, dag?m+4:m);
  return 0;
}

/*
 * Type 22: Conditional Trap (aka, Reserved)
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 ----COND---
 *
 */
unsigned int prettyop_trap(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_trap]");
  return 0;
}

/*
 * Type 23: DIVQ
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  1  1  0  0  0  0  0 ---XOP--  0  0  0  0  0  0  0  0
 *
 */
unsigned int prettyop_divq(unsigned int opcode, char *buf)
{
  unsigned int xop = (opcode >> 8) & 0x7;
  sprintf(buf, "divq %s", axops[xop]);
  return 0;
}

/*
 * Type 24: DIVS
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  1  0  0  0  0 -YOP- -XOP-  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int prettyop_divs(unsigned int opcode, char *buf)
{
  unsigned int yop = (opcode >> 11) & 0x3;
  unsigned int xop = (opcode >> 8) & 0x7;
  sprintf(buf, "divs %s, %s", ayops[yop], axops[xop]);
  return 0;
}

/*
 * Type 25: Saturate MR
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int prettyop_satmr(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_satmr]");
  return 0;
}

/*
 * Type 26: Stack Control
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0 PP LP CP -SPP-
 *
 */
unsigned int prettyop_stackctl(unsigned int opcode, char *buf)
{
  sprintf(buf, "[prettyop_stackctl]");
  return 0;
}

/*
 * Type 27: Call/Jump on Flag In
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  1 ------------Address1--------------- -Ad2- -F -S
 *
 *  Address1 - 12LSBs
 *  Address2 -  2MSBs
 *         F - FIC
 *
 */
unsigned int prettyop_jumpfi(unsigned int opcode, char *buf)
{
  unsigned int addr = (opcode >> 4) & 0xfff;
  unsigned int addrmsb = (opcode >> 2) & 0x3;
  unsigned int fic = (opcode >> 1) & 0x1;
  unsigned int s = (opcode & 0x1);
  sprintf(buf, "if %sfi %s 0x%04x",
	  fic?"":"not ",
	  s?"call":"jump",
	  (addrmsb<<12)+addr);
  
  return 0;
}

/*
 * Type 28: Modify Flag Out
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  0  0  0  0 --F2- --F1- --F0- --FO- ---COND----
 *  
 *  F2 - FL2 
 *  F1 - FL1
 *  F0 - FL0
 *  FO - FLAG_OUT
 *
 */
unsigned int prettyop_modifyfl(unsigned int opcode, char *buf)
{
  unsigned int fl[4] = {
    (opcode >> 10) & 0x3,
    (opcode >> 8) & 0x3,
    (opcode >> 6) & 0x3,
    (opcode >> 4) & 0x3,
  };
  unsigned int cond = opcode & 0xf;
  int cur = 0, i = 0;
  
  if (cond < 0xf)
    {
      sprintf(buf, "if %s ", conds[cond]);
      i += 6;
    }
  for (cur=0;cur<3;cur++)
    {
      if (fl[cur])
	{
	  sprintf(&(buf[i]), "%s fl%d ",
		  flagops[fl[cur]],
		  cur);
	  i += strlen(flagops[fl[cur]]) + 5;
	}
    }
  if (fl[3])
    {
      sprintf(&(buf[i]), "%s flag_out", 
	      flagops[fl[3]]);
    }
  return 0;
}

/*
 * Type 29: IO Memory Space R/W
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  0  1 -D -----------ADDR----------------- ---DREG----
 *
 */
unsigned int prettyop_iorw(unsigned int opcode, char *buf)
{
  unsigned int addr = (opcode >> 4) & 0x7ff;
  unsigned int dreg = opcode & 0xf;
  unsigned int dir = opcode & BIT15;
  if (dir)
    sprintf(buf, "io(%d) = %s", addr, regs[dreg][0]);
  else
    sprintf(buf, "%s = io(%d)", regs[dreg][0], addr);
  return 0;
}

/*
 * Type 30: NOP
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int prettyop_nop(unsigned int opcode, char *buf)
{
  sprintf(buf, "nop");
  return 0;
}

/*
 * Type 31: Idle
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
 *
 */
unsigned int prettyop_idle(unsigned int opcode, char *buf)
{
  sprintf(buf, "idle");
  return 0;
}

/*
 * Type 32: Idle(n) [Slow Idle]
 *
 * 23 22 21 20 19 18 17 16 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  0  0  0  0  0  0  1  0  1  0  0  0  0  0  0  0  0  0  0  0 ---DV------
 *
 */
unsigned int prettyop_idlen(unsigned int opcode, char *buf)
{
  sprintf(buf, "idle(%d)", opcode & 0xf);
  return 0;
}
